%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Author: Renan Weege Achjian      May 5, 2018     Version: 1.0   %
%                                                                 %
% Description:                                                    %
% Neural network with a single neuron (linear regression)         %
% D = 1 inputs, one output, uses gradient descent (LMS algorythm) %
% Batch training                                                  %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% Eta = 0.05
eta = 0.05;
D = 1;               % Dimension: number of inputs
learningSteps = 70; % Number of learning epochs

x = [-0.5 -0.2 -0.1 0.3 0.4 0.5 0.7]; % Input training data (x1 = first line, x2 = second line, etc.)
d = [-1    1    2   3.2 3.5 5   6  ]; % Expected output
N = size(x, 2);                       % Number of training samples

w = [0; 0]; % Initial parameters, in the format: [b; w1; w2; ...; wD]

figure(1)
hold on
grid on
plot(x(1,:), d, '*k'); % Data points
hdl2 = plot(x, [ones(1,N); x]' * w, '-r'); % Calculated mean line
title('Training data and learned mean line')

figure(2)
hold on
grid on
title('Learning curve')
xlabel('Epoch')
ylabel('Mean Squared Error')

figure(3)
hold on
grid on
title('Weights')

for k = 1:learningSteps
    sumEi = zeros(D+1,1);  % Sum of errors for the whole training dataset
    for i = 1:N
        currentSample = i; % Choose the next sample
        inputVector = [1; x(:,currentSample)];
    
        yi = dot(w, inputVector); % yi shoud be as close as possible to d(currentSample)
        sumEi = sumEi + (d(currentSample) - yi) * inputVector; % The error we are trying to minimize
    end
    
    w = w + eta * sumEi;     % LMS algorythm. Calculate next values of w, minimizing the error
    y = [ones(1,N); x]' * w; % The new mean line, calculated for all samples
    
    J = (1/(2*N)) * sum((d' - y).^2); % The mean squared error
    
    figure(1)
    delete(hdl2) % Erase previous mean line
    hdl2 = plot(x, y, '-r'); % Plot the calculated mean line
    
    figure(2)
    plot(k, J, '-+r')
    
    figure(3)
    plot(k, w(1), '-+b')
    plot(k, w(2), '-+r')
    
    pause(0.3)
end

%% Eta = 0.1
eta = 0.1;

w = [0; 0]; % Initial parameters, in the format: [b; w1; w2; ...; wD]

hdl2 = plot(x, [ones(1,N); x]' * w, '-g'); % Calculated mean line

for k = 1:learningSteps
    sumEi = zeros(D+1,1);  % Sum of errors for the whole training dataset
    for i = 1:N
        currentSample = i; % Choose the next sample
        inputVector = [1; x(:,currentSample)];
    
        yi = dot(w, inputVector); % yi shoud be as close as possible to d(currentSample)
        sumEi = sumEi + (d(currentSample) - yi) * inputVector; % The error we are trying to minimize
    end
    
    w = w + eta * sumEi;     % LMS algorythm. Calculate next values of w, minimizing the error
    y = [ones(1,N); x]' * w; % The new mean line, calculated for all samples
    
    J = (1/(2*N)) * sum((d' - y).^2); % The mean squared error
    
    figure(1)
    delete(hdl2) % Erase previous mean line
    hdl2 = plot(x, y, '-g'); % Plot the calculated mean line
    
    figure(2)
    plot(k, J, '-+g')
    
    figure(3)
    plot(k, w(1), '-+c')
    plot(k, w(2), '-+g')
    
    pause(0.3)
end

%% Eta = 1
eta = 1;

w = [0; 0]; % Initial parameters, in the format: [b; w1; w2; ...; wD]

hdl2 = plot(x, [ones(1,N); x]' * w, '-g'); % Calculated mean line

for k = 1:learningSteps
    sumEi = zeros(D+1,1);  % Sum of errors for the whole training dataset
    for i = 1:N
        currentSample = i; % Choose the next sample
        inputVector = [1; x(:,currentSample)];
    
        yi = dot(w, inputVector); % yi shoud be as close as possible to d(currentSample)
        sumEi = sumEi + (d(currentSample) - yi) * inputVector; % The error we are trying to minimize
    end
    
    w = w + eta * sumEi;     % LMS algorythm. Calculate next values of w, minimizing the error
    y = [ones(1,N); x]' * w; % The new mean line, calculated for all samples
    
    J = (1/(2*N)) * sum((d' - y).^2); % The mean squared error
    
    figure(1)
    delete(hdl2) % Erase previous mean line
    hdl2 = plot(x, y, '-y'); % Plot the calculated mean line
    
    figure(2)
    plot(k, J, '-+y')
    
    figure(3)
    plot(k, w(1), '-+m')
    plot(k, w(2), '-+y')
    
    pause(0.3)
end

figure(2)
legend(['Mean squared error - \eta = ' num2str(eta)])

figure(3)
legend('b', 'w1')