%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Author: Renan Weege Achjian      May 5, 2018     Version: 1.0   %
%                                                                 %
% Description:                                                    %
% Neural network with a single neuron (linear regression)         %
% D = 2 inputs, one output, uses gradient descent (LMS algorythm) %
% Sample by sample training                                       %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% Eta = 0.01
eta = 0.01;
D = 2;               % Dimension: number of inputs
learningSteps = 210; % Number of learning epochs

x = [-0.5 -0.2 -0.1 0.3 0.4 0.5 0.7; 3 3 2.5 2 -1 -1 -4]; % Input training data (x1 = first line, x2 = second line, etc.)
d = [-3   -1    0   1.2 1.5 3   4  ]; % Expected output
N = size(x, 2);                       % Number of training samples

w = [0; 0; 0]; % Initial parameters, in the format: [b; w1; w2; ...; wD]

figure(1)
hold on
grid on
plot3(x(1,:), x(2,:), d, '*k'); % Data points
hdl2 = patch(x(1,:), x(2,:), [ones(1,N); x]' * w, '+r'); % Calculated mean plane
hdl2.FaceVertexAlphaData = 0.2;
hdl2.FaceAlpha = 'flat' ;
title('Training data and learned mean plane')

figure(2)
hold on
grid on
title('Learning curve')
xlabel('Epoch')
ylabel('Mean Squared Error')

figure(3)
hold on
grid on
title('Weights')

for k = 1:learningSteps
    currentSample = randi(N, 1); % Choose a sample randomly
    inputVector = [1; x(:,currentSample)];
    
    yi = dot(w, inputVector);   % yi shoud be as close as possible to d(currentSample)
    Ei = d(currentSample) - yi; % The error we are trying to minimize
    
    w = w + eta * Ei * inputVector; % LMS algorythm. Calculate next values of w, minimizing E
    y = [ones(1,N); x]' * w;        % The new mean plane, calculated for all samples
    
    J = (1/(2*N)) * sum((d' - y).^2); % The mean squared error
    
    figure(1)
    hdl = plot3(x(1,currentSample), x(2,currentSample), d(currentSample), 'og'); % Plot the point considered in this iteration
    delete(hdl2) % Erase previous mean line
    hdl2 = patch(x(1,:), x(2,:), y, '+r'); % Plot the calculated mean plane
    hdl2.FaceVertexAlphaData = 0.2;
    hdl2.FaceAlpha = 'flat' ;
    
    figure(2)
    plot(k, J, '-+r')
    
    figure(3)
    plot(k, w(1), '-+b')
    plot(k, w(2), '-+r')
    plot(k, w(3), '-+c')
    
    pause(0.2)
    
    delete(hdl)
end

%% Eta = 0.1
eta = 0.1;

w = [0; 0; 0]; % Initial parameters, in the format: [b; w1; w2; ...; wD]

hdl2 = patch(x(1,:), x(2,:), [ones(1,N); x]' * w, '+g'); % Calculated mean plane
hdl2.FaceVertexAlphaData = 0.2;
hdl2.FaceAlpha = 'flat' ;

for k = 1:learningSteps
    currentSample = randi(N, 1); % Choose a sample randomly
    inputVector = [1; x(:,currentSample)];
    
    yi = dot(w, inputVector);   % yi shoud be as close as possible to d(currentSample)
    Ei = d(currentSample) - yi; % The error we are trying to minimize
    
    w = w + eta * Ei * inputVector; % LMS algorythm. Calculate next values of w, minimizing E
    y = [ones(1,N); x]' * w;        % The new mean line, calculated for all samples
    
    J = (1/(2*N)) * sum((d' - y).^2); % The mean squared error
    
    figure(1)
    hdl = plot3(x(1,currentSample), x(2,currentSample), d(currentSample), 'og'); % Plot the point considered in this iteration
    delete(hdl2) % Erase previous mean line
    hdl2 = patch(x(1,:), x(2,:), y, '+g'); % Plot the calculated mean plane
    hdl2.FaceVertexAlphaData = 0.2;
    
    figure(2)
    plot(k, J, '-+g')
    
    figure(3)
    plot(k, w(1), '-+m')
    plot(k, w(2), '-+g')
    plot(k, w(3), '-+y')
    
    pause(0.2)
    
    delete(hdl)
end

%% Eta = 0.5
eta = 0.5;

w = [0; 0; 0]; % Initial parameters, in the format: [b; w1; w2; ...; wD]

hdl2 = patch(x(1,:), x(2,:), [ones(1,N); x]' * w, '+g'); % Calculated mean plane
hdl2.FaceVertexAlphaData = 0.2;
hdl2.FaceAlpha = 'flat' ;

for k = 1:learningSteps
    currentSample = randi(N, 1); % Choose a sample randomly
    inputVector = [1; x(:,currentSample)];
    
    yi = dot(w, inputVector);   % yi shoud be as close as possible to d(currentSample)
    Ei = d(currentSample) - yi; % The error we are trying to minimize
    
    w = w + eta * Ei * inputVector; % LMS algorythm. Calculate next values of w, minimizing E
    y = [ones(1,N); x]' * w;        % The new mean line, calculated for all samples
    
    J = (1/(2*N)) * sum((d' - y).^2); % The mean squared error
    
    figure(1)
    hdl = plot3(x(1,currentSample), x(2,currentSample), d(currentSample), 'og'); % Plot the point considered in this iteration
    delete(hdl2) % Erase previous mean line
    hdl2 = patch(x(1,:), x(2,:), y, '+c'); % Plot the calculated mean plane
    hdl2.FaceVertexAlphaData = 0.2;
    
    figure(2)
    plot(k, J, '-+c')
    
    figure(3)
    plot(k, w(1), '-+c')
    plot(k, w(2), '-+k')
    plot(k, w(3), '-+r')
    
    pause(0.2)
    
    delete(hdl)
end

figure(2)
legend(['Mean squared error - \eta = ' num2str(eta)])

figure(3)
legend('b', 'w1')